@@ -21,26 +21,31 @@
 #include <arpa/inet.h>
 #include <string.h>
 #include <linux/if.h>
+#include <linux/if_ether.h>
 
 #include "utils.h"
 #include "tc_util.h"
 
+extern int show_pretty;
+
 static void explain(void)
 {
-	fprintf(stderr, "Usage: ... u32 [ match SELECTOR ... ] [ link HTID ] [ classid CLASSID ]\n");
-	fprintf(stderr, "               [ police POLICE_SPEC ] [ offset OFFSET_SPEC ]\n");
+	fprintf(stderr, "Usage: ... u32 [ match SELECTOR ... ] [ link HTID ]"
+		" [ classid CLASSID ]\n");
+	fprintf(stderr, "               [ action ACTION_SPEC ]"
+		" [ offset OFFSET_SPEC ]\n");
 	fprintf(stderr, "               [ ht HTID ] [ hashkey HASHKEY_SPEC ]\n");
 	fprintf(stderr, "               [ sample SAMPLE ]\n");
 	fprintf(stderr, "or         u32 divisor DIVISOR\n");
 	fprintf(stderr, "\n");
 	fprintf(stderr, "Where: SELECTOR := SAMPLE SAMPLE ...\n");
-	fprintf(stderr, "       SAMPLE := { ip | ip6 | udp | tcp | icmp | u{32|16|8} | mark } SAMPLE_ARGS\n");
+	fprintf(stderr, "       SAMPLE := { ip | ip6 | udp | tcp | icmp |"
+		" u{32|16|8} | mark } SAMPLE_ARGS [divisor DIVISOR]\n");
 	fprintf(stderr, "       FILTERID := X:Y:Z\n");
+	fprintf(stderr, "\nNOTE: CLASSID is parsed at hexadecimal input.\n");
 }
 
-#define usage() return(-1)
-
-int get_u32_handle(__u32 *handle, char *str)
+static int get_u32_handle(__u32 *handle, const char *str)
 {
 	__u32 htid=0, hash=0, nodeid=0;
 	char *tmp = strchr(str, ':');
@@ -75,7 +80,7 @@ int get_u32_handle(__u32 *handle, char *str)
 	return 0;
 }
 
-char * sprint_u32_handle(__u32 handle, char *buf)
+static char * sprint_u32_handle(__u32 handle, char *buf)
 {
 	int bsize = SPRINT_BSIZE-1;
 	__u32 htid = TC_U32_HTID(handle);
@@ -109,7 +114,8 @@ char * sprint_u32_handle(__u32 handle, char *buf)
 	return buf;
 }
 
-static int pack_key(struct tc_u32_sel *sel, __u32 key, __u32 mask, int off, int offmask)
+static int pack_key(struct tc_u32_sel *sel, __u32 key, __u32 mask,
+		    int off, int offmask)
 {
 	int i;
 	int hwm = sel->nkeys;
@@ -140,14 +146,16 @@ static int pack_key(struct tc_u32_sel *sel, __u32 key, __u32 mask, int off, int
 	return 0;
 }
 
-static int pack_key32(struct tc_u32_sel *sel, __u32 key, __u32 mask, int off, int offmask)
+static int pack_key32(struct tc_u32_sel *sel, __u32 key, __u32 mask,
+		      int off, int offmask)
 {
 	key = htonl(key);
 	mask = htonl(mask);
 	return pack_key(sel, key, mask, off, offmask);
 }
 
-static int pack_key16(struct tc_u32_sel *sel, __u32 key, __u32 mask, int off, int offmask)
+static int pack_key16(struct tc_u32_sel *sel, __u32 key, __u32 mask,
+		      int off, int offmask)
 {
 	if (key > 0xFFFF || mask > 0xFFFF)
 		return -1;
@@ -186,7 +194,7 @@ static int pack_key8(struct tc_u32_sel *sel, __u32 key, __u32 mask, int off, int
 }
 
 
-int parse_at(int *argc_p, char ***argv_p, int *off, int *offmask)
+static int parse_at(int *argc_p, char ***argv_p, int *off, int *offmask)
 {
 	int argc = *argc_p;
 	char **argv = *argv_p;
@@ -215,7 +223,8 @@ int parse_at(int *argc_p, char ***argv_p, int *off, int *offmask)
 }
 
 
-static int parse_u32(int *argc_p, char ***argv_p, struct tc_u32_sel *sel, int off, int offmask)
+static int parse_u32(int *argc_p, char ***argv_p, struct tc_u32_sel *sel,
+		     int off, int offmask)
 {
 	int res = -1;
 	int argc = *argc_p;
@@ -246,7 +255,8 @@ static int parse_u32(int *argc_p, char ***argv_p, struct tc_u32_sel *sel, int of
 	return res;
 }
 
-static int parse_u16(int *argc_p, char ***argv_p, struct tc_u32_sel *sel, int off, int offmask)
+static int parse_u16(int *argc_p, char ***argv_p, struct tc_u32_sel *sel,
+		     int off, int offmask)
 {
 	int res = -1;
 	int argc = *argc_p;
@@ -276,7 +286,8 @@ static int parse_u16(int *argc_p, char ***argv_p, struct tc_u32_sel *sel, int of
 	return res;
 }
 
-static int parse_u8(int *argc_p, char ***argv_p, struct tc_u32_sel *sel, int off, int offmask)
+static int parse_u8(int *argc_p, char ***argv_p, struct tc_u32_sel *sel,
+		    int off, int offmask)
 {
 	int res = -1;
 	int argc = *argc_p;
@@ -310,7 +321,8 @@ static int parse_u8(int *argc_p, char ***argv_p, struct tc_u32_sel *sel, int off
 	return res;
 }
 
-static int parse_ip_addr(int *argc_p, char ***argv_p, struct tc_u32_sel *sel, int off)
+static int parse_ip_addr(int *argc_p, char ***argv_p, struct tc_u32_sel *sel,
+			 int off)
 {
 	int res = -1;
 	int argc = *argc_p;
@@ -344,7 +356,8 @@ static int parse_ip_addr(int *argc_p, char ***argv_p, struct tc_u32_sel *sel, in
 	return res;
 }
 
-static int parse_ip6_addr(int *argc_p, char ***argv_p, struct tc_u32_sel *sel, int off)
+static int parse_ip6_addr(int *argc_p, char ***argv_p,
+			  struct tc_u32_sel *sel, int off)
 {
 	int res = -1;
 	int argc = *argc_p;
@@ -370,12 +383,16 @@ static int parse_ip6_addr(int *argc_p, char ***argv_p, struct tc_u32_sel *sel, i
 	plen = addr.bitlen;
 	for (i=0; i<plen; i+=32) {
 //		if (((i+31)&~0x1F)<=plen) {
-		if (((i+31))<=plen) {
-			if ((res = pack_key(sel, addr.data[i/32], 0xFFFFFFFF, off+4*(i/32), offmask)) < 0)
+		if (i + 31 <= plen) {
+			res = pack_key(sel, addr.data[i/32],
+				       0xFFFFFFFF, off+4*(i/32), offmask);
+			if (res < 0)
 				return -1;
-		} else if (i<plen) {
-			__u32 mask = htonl(0xFFFFFFFF<<(32-(plen-i)));
-			if ((res = pack_key(sel, addr.data[i/32], mask, off+4*(i/32), offmask)) < 0)
+		} else if (i < plen) {
+			__u32 mask = htonl(0xFFFFFFFF << (32 - (plen -i )));
+			res = pack_key(sel, addr.data[i/32],
+				       mask, off+4*(i/32), offmask);
+			if (res < 0)
 				return -1;
 		}
 	}
@@ -386,6 +403,83 @@ static int parse_ip6_addr(int *argc_p, char ***argv_p, struct tc_u32_sel *sel, i
 	return res;
 }
 
+static int parse_ip6_class(int *argc_p, char ***argv_p, struct tc_u32_sel *sel)
+{
+	int res = -1;
+	int argc = *argc_p;
+	char **argv = *argv_p;
+	__u32 key;
+	__u32 mask;
+	int off = 0;
+	int offmask = 0;
+
+	if (argc < 2)
+		return -1;
+
+	if (get_u32(&key, *argv, 0))
+		return -1;
+	argc--; argv++;
+
+	if (get_u32(&mask, *argv, 16))
+		return -1;
+	argc--; argv++;
+
+	if (key > 0xFF || mask > 0xFF)
+		return -1;
+
+	key <<= 20;
+	mask <<= 20;
+	key = htonl(key);
+	mask = htonl(mask);
+
+	res = pack_key(sel, key, mask, off, offmask);
+	if (res < 0)
+		return -1;
+
+	*argc_p = argc;
+	*argv_p = argv;
+	return 0;
+}
+
+static int parse_ether_addr(int *argc_p, char ***argv_p,
+			    struct tc_u32_sel *sel, int off)
+{
+	int res = -1;
+	int argc = *argc_p;
+	char **argv = *argv_p;
+	__u8 addr[6];
+	int offmask = 0;
+	int i;
+
+	if (argc < 1)
+		return -1;
+
+	if (sscanf(*argv, "%hhx:%hhx:%hhx:%hhx:%hhx:%hhx",
+		   addr + 0, addr + 1, addr + 2,
+		   addr + 3, addr + 4, addr + 5) != 6) {
+		fprintf(stderr, "parse_ether_addr: improperly formed address '%s'\n",
+			*argv);
+		return -1;
+	}
+
+	argc--; argv++;
+	if (argc > 0 && strcmp(argv[0], "at") == 0) {
+		NEXT_ARG();
+		if (parse_at(&argc, &argv, &off, &offmask))
+			return -1;
+	}
+
+	for (i = 0; i < 6; i++) {
+		res = pack_key8(sel, addr[i], 0xFF, off + i, offmask);
+		if (res < 0)
+			return -1;
+	}
+
+	*argc_p = argc;
+	*argv_p = argv;
+	return res;
+}
+
 static int parse_ip(int *argc_p, char ***argv_p, struct tc_u32_sel *sel)
 {
 	int res = -1;
@@ -398,82 +492,54 @@ static int parse_ip(int *argc_p, char ***argv_p, struct tc_u32_sel *sel)
 	if (strcmp(*argv, "src") == 0) {
 		NEXT_ARG();
 		res = parse_ip_addr(&argc, &argv, sel, 12);
-		goto done;
-	}
-	if (strcmp(*argv, "dst") == 0) {
+	} else if (strcmp(*argv, "dst") == 0) {
 		NEXT_ARG();
 		res = parse_ip_addr(&argc, &argv, sel, 16);
-		goto done;
-	}
-	if (strcmp(*argv, "tos") == 0 ||
+	} else if (strcmp(*argv, "tos") == 0 ||
 	    matches(*argv, "dsfield") == 0) {
 		NEXT_ARG();
 		res = parse_u8(&argc, &argv, sel, 1, 0);
-		goto done;
-	}
-	if (strcmp(*argv, "ihl") == 0) {
+	} else if (strcmp(*argv, "ihl") == 0) {
 		NEXT_ARG();
 		res = parse_u8(&argc, &argv, sel, 0, 0);
-		goto done;
-	}
-	if (strcmp(*argv, "protocol") == 0) {
+	} else if (strcmp(*argv, "protocol") == 0) {
 		NEXT_ARG();
 		res = parse_u8(&argc, &argv, sel, 9, 0);
-		goto done;
-	}
-	if (matches(*argv, "precedence") == 0) {
+	} else if (matches(*argv, "precedence") == 0) {
 		NEXT_ARG();
 		res = parse_u8(&argc, &argv, sel, 1, 0);
-		goto done;
-	}
-	if (strcmp(*argv, "nofrag") == 0) {
+	} else if (strcmp(*argv, "nofrag") == 0) {
 		argc--; argv++;
 		res = pack_key16(sel, 0, 0x3FFF, 6, 0);
-		goto done;
-	}
-	if (strcmp(*argv, "firstfrag") == 0) {
+	} else if (strcmp(*argv, "firstfrag") == 0) {
 		argc--; argv++;
-		res = pack_key16(sel, 0, 0x1FFF, 6, 0);
-		goto done;
-	}
-	if (strcmp(*argv, "df") == 0) {
+		res = pack_key16(sel, 0x2000, 0x3FFF, 6, 0);
+	} else if (strcmp(*argv, "df") == 0) {
 		argc--; argv++;
 		res = pack_key16(sel, 0x4000, 0x4000, 6, 0);
-		goto done;
-	}
-	if (strcmp(*argv, "mf") == 0) {
+	} else if (strcmp(*argv, "mf") == 0) {
 		argc--; argv++;
 		res = pack_key16(sel, 0x2000, 0x2000, 6, 0);
-		goto done;
-	}
-	if (strcmp(*argv, "dport") == 0) {
+	} else if (strcmp(*argv, "dport") == 0) {
 		NEXT_ARG();
 		res = parse_u16(&argc, &argv, sel, 22, 0);
-		goto done;
-	}
-	if (strcmp(*argv, "sport") == 0) {
+	} else if (strcmp(*argv, "sport") == 0) {
 		NEXT_ARG();
 		res = parse_u16(&argc, &argv, sel, 20, 0);
-		goto done;
-	}
-	if (strcmp(*argv, "icmp_type") == 0) {
+	} else if (strcmp(*argv, "icmp_type") == 0) {
 		NEXT_ARG();
 		res = parse_u8(&argc, &argv, sel, 20, 0);
-		goto done;
-	}
-	if (strcmp(*argv, "icmp_code") == 0) {
+	} else if (strcmp(*argv, "icmp_code") == 0) {
 		NEXT_ARG();
-		res = parse_u8(&argc, &argv, sel, 20, 1);
-		goto done;
-	}
-	return -1;
+		res = parse_u8(&argc, &argv, sel, 21, 0);
+	} else
+		return -1;
 
-done:
 	*argc_p = argc;
 	*argv_p = argv;
 	return res;
 }
-
+				
 static int parse_ip6(int *argc_p, char ***argv_p, struct tc_u32_sel *sel)
 {
 	int res = -1;
@@ -486,51 +552,58 @@ static int parse_ip6(int *argc_p, char ***argv_p, struct tc_u32_sel *sel)
 	if (strcmp(*argv, "src") == 0) {
 		NEXT_ARG();
 		res = parse_ip6_addr(&argc, &argv, sel, 8);
-		goto done;
-	}
-	if (strcmp(*argv, "dst") == 0) {
+	} else if (strcmp(*argv, "dst") == 0) {
 		NEXT_ARG();
 		res = parse_ip6_addr(&argc, &argv, sel, 24);
-		goto done;
-	}
-	if (strcmp(*argv, "priority") == 0) {
+	} else if (strcmp(*argv, "priority") == 0) {
 		NEXT_ARG();
-		res = parse_u8(&argc, &argv, sel, 0, 0);
-		goto done;
-	}
-	if (strcmp(*argv, "protocol") == 0) {
+		res = parse_ip6_class(&argc, &argv, sel);
+	} else if (strcmp(*argv, "protocol") == 0) {
 		NEXT_ARG();
 		res = parse_u8(&argc, &argv, sel, 6, 0);
-		goto done;
-	}
-	if (strcmp(*argv, "flowlabel") == 0) {
+	} else if (strcmp(*argv, "flowlabel") == 0) {
 		NEXT_ARG();
 		res = parse_u32(&argc, &argv, sel, 0, 0);
-		goto done;
-	}
-	if (strcmp(*argv, "dport") == 0) {
+	} else if (strcmp(*argv, "dport") == 0) {
 		NEXT_ARG();
 		res = parse_u16(&argc, &argv, sel, 42, 0);
-		goto done;
-	}
-	if (strcmp(*argv, "sport") == 0) {
+	} else if (strcmp(*argv, "sport") == 0) {
 		NEXT_ARG();
 		res = parse_u16(&argc, &argv, sel, 40, 0);
-		goto done;
-	}
-	if (strcmp(*argv, "icmp_type") == 0) {
+	} else if (strcmp(*argv, "icmp_type") == 0) {
 		NEXT_ARG();
 		res = parse_u8(&argc, &argv, sel, 40, 0);
-		goto done;
-	}
-	if (strcmp(*argv, "icmp_code") == 0) {
+	} else if (strcmp(*argv, "icmp_code") == 0) {
 		NEXT_ARG();
 		res = parse_u8(&argc, &argv, sel, 41, 1);
-		goto done;
+	} else
+		return -1;
+
+	*argc_p = argc;
+	*argv_p = argv;
+	return res;
+}
+
+static int parse_ether(int *argc_p, char ***argv_p, struct tc_u32_sel *sel)
+{
+	int res = -1;
+	int argc = *argc_p;
+	char **argv = *argv_p;
+
+	if (argc < 2)
+		return -1;
+
+	if (strcmp(*argv, "src") == 0) {
+		NEXT_ARG();
+		res = parse_ether_addr(&argc, &argv, sel, -8);
+	} else if (strcmp(*argv, "dst") == 0) {
+		NEXT_ARG();
+		res = parse_ether_addr(&argc, &argv, sel, -14);
+	} else {
+		fprintf(stderr, "Unknown match: ether %s\n", *argv);
+		return -1;
 	}
-	return -1;
 
-done:
 	*argc_p = argc;
 	*argv_p = argv;
 	return res;
@@ -549,21 +622,18 @@ static int parse_udp(int *argc_p, char ***argv_p, struct tc_u32_sel *sel)
 	if (strcmp(*argv, "src") == 0) {
 		NEXT_ARG();
 		res = parse_u16(&argc, &argv, sel, 0, -1);
-		goto done;
-	}
-	if (strcmp(*argv, "dst") == 0) {
+	} else if (strcmp(*argv, "dst") == 0) {
 		NEXT_ARG();
 		res = parse_u16(&argc, &argv, sel, 2, -1);
-		goto done;
-	}
-	return -1;
+	} else
+		return -1;
 
-done:
 	*argc_p = argc;
 	*argv_p = argv;
 	return res;
 }
 
+
 static int parse_icmp(int *argc_p, char ***argv_p, struct tc_u32_sel *sel)
 {
 	int res = -1;
@@ -576,16 +646,12 @@ static int parse_icmp(int *argc_p, char ***argv_p, struct tc_u32_sel *sel)
 	if (strcmp(*argv, "type") == 0) {
 		NEXT_ARG();
 		res = parse_u8(&argc, &argv, sel, 0, -1);
-		goto done;
-	}
-	if (strcmp(*argv, "code") == 0) {
+	} else if (strcmp(*argv, "code") == 0) {
 		NEXT_ARG();
 		res = parse_u8(&argc, &argv, sel, 1, -1);
-		goto done;
-	}
-	return -1;
+	} else
+		return -1;
 
-done:
 	*argc_p = argc;
 	*argv_p = argv;
 	return res;
@@ -626,7 +692,8 @@ static int parse_mark(int *argc_p, char ***argv_p, struct nlmsghdr *n)
 	return res;
 }
 
-static int parse_selector(int *argc_p, char ***argv_p, struct tc_u32_sel *sel, struct nlmsghdr *n)
+static int parse_selector(int *argc_p, char ***argv_p,
+			  struct tc_u32_sel *sel, struct nlmsghdr *n)
 {
 	int argc = *argc_p;
 	char **argv = *argv_p;
@@ -638,52 +705,36 @@ static int parse_selector(int *argc_p, char ***argv_p, struct tc_u32_sel *sel, s
 	if (matches(*argv, "u32") == 0) {
 		NEXT_ARG();
 		res = parse_u32(&argc, &argv, sel, 0, 0);
-		goto done;
-	}
-	if (matches(*argv, "u16") == 0) {
+	} else if (matches(*argv, "u16") == 0) {
 		NEXT_ARG();
 		res = parse_u16(&argc, &argv, sel, 0, 0);
-		goto done;
-	}
-	if (matches(*argv, "u8") == 0) {
+	} else if (matches(*argv, "u8") == 0) {
 		NEXT_ARG();
 		res = parse_u8(&argc, &argv, sel, 0, 0);
-		goto done;
-	}
-	if (matches(*argv, "ip") == 0) {
+	} else if (matches(*argv, "ip") == 0) {
 		NEXT_ARG();
 		res = parse_ip(&argc, &argv, sel);
-		goto done;
-	}
-	if (matches(*argv, "ip6") == 0) {
+	} else 	if (matches(*argv, "ip6") == 0) {
 		NEXT_ARG();
 		res = parse_ip6(&argc, &argv, sel);
-		goto done;
-	}
-	if (matches(*argv, "udp") == 0) {
+	} else if (matches(*argv, "udp") == 0) {
 		NEXT_ARG();
 		res = parse_udp(&argc, &argv, sel);
-		goto done;
-	}
-	if (matches(*argv, "tcp") == 0) {
+	} else if (matches(*argv, "tcp") == 0) {
 		NEXT_ARG();
 		res = parse_tcp(&argc, &argv, sel);
-		goto done;
-	}
-	if (matches(*argv, "icmp") == 0) {
+	} else if (matches(*argv, "icmp") == 0) {
 		NEXT_ARG();
 		res = parse_icmp(&argc, &argv, sel);
-		goto done;
-	}
-	if (matches(*argv, "mark") == 0) {
+	} else if (matches(*argv, "mark") == 0) {
 		NEXT_ARG();
 		res = parse_mark(&argc, &argv, n);
-		goto done;
-	}
+	} else if (matches(*argv, "ether") == 0) {
+		NEXT_ARG();
+		res = parse_ether(&argc, &argv, sel);
+	} else 
+		return -1;
 
-	return -1;
-
-done:
 	*argc_p = argc;
 	*argv_p = argv;
 	return res;
@@ -771,7 +822,158 @@ static int parse_hashkey(int *argc_p, char ***argv_p, struct tc_u32_sel *sel)
 	return 0;
 }
 
-static int u32_parse_opt(struct filter_util *qu, char *handle, int argc, char **argv, struct nlmsghdr *n)
+static void print_ipv4(FILE *f, const struct tc_u32_key *key)
+{
+	char abuf[256];
+
+	switch (key->off) {
+	case 0:
+		switch (ntohl(key->mask)) {
+		case 0x0f000000:
+			fprintf(f, "\n  match IP ihl %u", ntohl(key->val) >> 24);
+			return;
+		case 0x00ff0000:
+			fprintf(f, "\n  match IP dsfield %#x", ntohl(key->val) >> 16);
+			return;
+		}
+		break;
+	case 8:
+		if (ntohl(key->mask) == 0x00ff0000) {
+			fprintf(f, "\n  match IP protocol %d", ntohl(key->val) >> 16);
+			return;
+		}
+		break;
+	case 12:
+	case 16: {
+			int bits = mask2bits(key->mask);
+			if (bits >= 0) {
+				fprintf(f, "\n  %s %s/%d", 
+					key->off == 12 ? "match IP src" : "match IP dst",
+					inet_ntop(AF_INET, &key->val,
+						  abuf, sizeof(abuf)),
+					bits);
+				return;
+			}
+		}
+		break;
+
+	case 20:
+		switch (ntohl(key->mask)) {
+		case 0x0000ffff:
+			fprintf(f, "\n  match dport %u",
+				ntohl(key->val) & 0xffff);
+			return;
+		case 0xffff0000:
+			fprintf(f, "\n  match sport %u",
+				ntohl(key->val) >> 16);
+			return;
+		case 0xffffffff:
+			fprintf(f, "\n  match dport %u, match sport %u",
+				ntohl(key->val) & 0xffff,
+				ntohl(key->val) >> 16);
+
+			return;
+		}
+		/* XXX: Default print_raw */
+	}
+}
+
+static void print_ipv6(FILE *f, const struct tc_u32_key *key)
+{
+	char abuf[256];
+
+	switch (key->off) {
+	case 0:
+		switch (ntohl(key->mask)) {
+		case 0x0f000000:
+			fprintf(f, "\n  match IP ihl %u", ntohl(key->val) >> 24);
+			return;
+		case 0x00ff0000:
+			fprintf(f, "\n  match IP dsfield %#x", ntohl(key->val) >> 16);
+			return;
+		}
+		break;
+	case 8:
+		if (ntohl(key->mask) == 0x00ff0000) {
+			fprintf(f, "\n  match IP protocol %d", ntohl(key->val) >> 16);
+			return;
+		}
+		break;
+	case 12:
+	case 16: {
+			int bits = mask2bits(key->mask);
+			if (bits >= 0) {
+				fprintf(f, "\n  %s %s/%d", 
+					key->off == 12 ? "match IP src" : "match IP dst",
+					inet_ntop(AF_INET, &key->val,
+						  abuf, sizeof(abuf)),
+					bits);
+				return;
+			}
+		}
+		break;
+
+	case 20:
+		switch (ntohl(key->mask)) {
+		case 0x0000ffff:
+			fprintf(f, "\n  match sport %u",
+				ntohl(key->val) & 0xffff);
+			return;
+		case 0xffff0000:
+			fprintf(f, "\n  match dport %u",
+				ntohl(key->val) >> 16);
+			return;
+		case 0xffffffff:
+			fprintf(f, "\n  match sport %u, match dport %u",
+				ntohl(key->val) & 0xffff,
+				ntohl(key->val) >> 16);
+
+			return;
+		}
+		/* XXX: Default print_raw */
+	}
+}
+
+static void print_raw(FILE *f, const struct tc_u32_key *key)
+{
+	fprintf(f, "\n  match %08x/%08x at %s%d", 
+		(unsigned int)ntohl(key->val),
+		(unsigned int)ntohl(key->mask),
+		key->offmask ? "nexthdr+" : "",
+		key->off);
+}
+
+static const struct {
+	__u16 proto;
+	__u16 pad;
+	void (*pprinter)(FILE *f, const struct tc_u32_key *key);
+} u32_pprinters[] = {
+	{0, 	   0, print_raw},
+	{ETH_P_IP, 0, print_ipv4},
+	{ETH_P_IPV6, 0, print_ipv6},
+};
+
+static void show_keys(FILE *f, const struct tc_u32_key *key)
+{
+	int i = 0;
+
+	if (!show_pretty)
+		goto show_k;
+
+	for (i = 0; i < sizeof(u32_pprinters) / sizeof(u32_pprinters[0]); i++) {
+		if (u32_pprinters[i].proto == ntohs(f_proto)) {
+show_k:
+			u32_pprinters[i].pprinter(f, key);
+			return;
+		}
+	}
+
+	i = 0;
+	goto show_k;
+}
+
+static int u32_parse_opt(struct filter_util *qu, char *handle,
+			 int argc, char **argv, struct nlmsghdr *n)
 {
 	struct {
 		struct tc_u32_sel sel;
@@ -779,7 +981,7 @@ static int u32_parse_opt(struct filter_util *qu, char *handle, int argc, char **
 	} sel;
 	struct tcmsg *t = NLMSG_DATA(n);
 	struct rtattr *tail;
-	int sel_ok = 0;
+	int sel_ok = 0, terminal_ok = 0;
 	int sample_ok = 0;
 	__u32 htid = 0;
 	__u32 order = 0;
@@ -833,8 +1035,9 @@ static int u32_parse_opt(struct filter_util *qu, char *handle, int argc, char **
 		} else if (matches(*argv, "divisor") == 0) {
 			unsigned divisor;
 			NEXT_ARG();
-			if (get_unsigned(&divisor, *argv, 0) || divisor == 0 ||
-			    divisor > 0x100) {
+			if (get_unsigned(&divisor, *argv, 0) ||
+			    divisor == 0 ||
+			    divisor > 0x100 || ((divisor - 1) & divisor)) {
 				fprintf(stderr, "Illegal \"divisor\"\n");
 				return -1;
 			}
@@ -874,23 +1077,36 @@ static int u32_parse_opt(struct filter_util *qu, char *handle, int argc, char **
 				htid = (handle&0xFFFFF000);
 		} else if (strcmp(*argv, "sample") == 0) {
 			__u32 hash;
+			unsigned divisor = 0x100;
+
 			struct {
 				struct tc_u32_sel sel;
 				struct tc_u32_key keys[4];
 			} sel2;
+			memset(&sel2, 0, sizeof(sel2));
 			NEXT_ARG();
 			if (parse_selector(&argc, &argv, &sel2.sel, n)) {
 				fprintf(stderr, "Illegal \"sample\"\n");
 				return -1;
 			}
 			if (sel2.sel.nkeys != 1) {
-				fprintf(stderr, "\"sample\" must contain exactly ONE key.\n");
+				fprintf(stderr, "\"sample\" must contain"
+					" exactly ONE key.\n");
 				return -1;
 			}
+			if (*argv != 0 && strcmp(*argv, "divisor") == 0) {
+				NEXT_ARG();
+				if (get_unsigned(&divisor, *argv, 0) || divisor == 0 ||
+				    divisor > 0x100 || ((divisor - 1) & divisor)) {
+					fprintf(stderr, "Illegal sample \"divisor\"\n");
+					return -1;
+				}
+				NEXT_ARG();
+			}
 			hash = sel2.sel.keys[0].val&sel2.sel.keys[0].mask;
 			hash ^= hash>>16;
 			hash ^= hash>>8;
-			htid = ((hash<<12)&0xFF000)|(htid&0xFFF00000);
+			htid = ((hash%divisor)<<12)|(htid&0xFFF00000);
 			sample_ok = 1;
 			continue;
 		} else if (strcmp(*argv, "indev") == 0) {
@@ -911,6 +1127,7 @@ static int u32_parse_opt(struct filter_util *qu, char *handle, int argc, char **
 				fprintf(stderr, "Illegal \"action\"\n");
 				return -1;
 			}
+			terminal_ok++;
 			continue;
 
 		} else if (matches(*argv, "police") == 0) {
@@ -919,6 +1136,7 @@ static int u32_parse_opt(struct filter_util *qu, char *handle, int argc, char **
 				fprintf(stderr, "Illegal \"police\"\n");
 				return -1;
 			}
+			terminal_ok++;
 			continue;
 		} else if (strcmp(*argv, "help") == 0) {
 			explain();
@@ -931,6 +1149,10 @@ static int u32_parse_opt(struct filter_util *qu, char *handle, int argc, char **
 		argc--; argv++;
 	}
 
+	/* We dont necessarily need class/flowids */
+	if (terminal_ok)
+		sel.sel.flags |= TC_U32_TERMINAL;
+	
 	if (order) {
 		if (TC_U32_NODE(t->tcm_handle) && order != TC_U32_NODE(t->tcm_handle)) {
 			fprintf(stderr, "\"order\" contradicts \"handle\"\n");
@@ -942,12 +1164,14 @@ static int u32_parse_opt(struct filter_util *qu, char *handle, int argc, char **
 	if (htid)
 		addattr_l(n, MAX_MSG, TCA_U32_HASH, &htid, 4);
 	if (sel_ok)
-		addattr_l(n, MAX_MSG, TCA_U32_SEL, &sel, sizeof(sel.sel)+sel.sel.nkeys*sizeof(struct tc_u32_key));
+		addattr_l(n, MAX_MSG, TCA_U32_SEL, &sel, 
+			  sizeof(sel.sel)+sel.sel.nkeys*sizeof(struct tc_u32_key));
 	tail->rta_len = (void *) NLMSG_TAIL(n) - (void *) tail;
 	return 0;
 }
 
-static int u32_print_opt(struct filter_util *qu, FILE *f, struct rtattr *opt, __u32 handle)
+static int u32_print_opt(struct filter_util *qu, FILE *f, struct rtattr *opt,
+			 __u32 handle)
 {
 	struct rtattr *tb[TCA_U32_MAX+1];
 	struct tc_u32_sel *sel = NULL;
@@ -974,10 +1198,11 @@ static int u32_print_opt(struct filter_util *qu, FILE *f, struct rtattr *opt, __
 	}
 
 	if (tb[TCA_U32_DIVISOR]) {
-		fprintf(f, "ht divisor %d ", *(__u32*)RTA_DATA(tb[TCA_U32_DIVISOR]));
+		fprintf(f, "ht divisor %d ", rta_getattr_u32(tb[TCA_U32_DIVISOR]));
 	} else if (tb[TCA_U32_HASH]) {
-		__u32 htid = *(__u32*)RTA_DATA(tb[TCA_U32_HASH]);
-		fprintf(f, "key ht %x bkt %x ", TC_U32_USERHTID(htid), TC_U32_HASH(htid));
+		__u32 htid = rta_getattr_u32(tb[TCA_U32_HASH]);
+		fprintf(f, "key ht %x bkt %x ", TC_U32_USERHTID(htid),
+			TC_U32_HASH(htid));
 	} else {
 		fprintf(f, "??? ");
 	}
@@ -985,13 +1210,14 @@ static int u32_print_opt(struct filter_util *qu, FILE *f, struct rtattr *opt, __
 		SPRINT_BUF(b1);
 		fprintf(f, "%sflowid %s ",
 			!sel || !(sel->flags&TC_U32_TERMINAL) ? "*" : "",
-			sprint_tc_classid(*(__u32*)RTA_DATA(tb[TCA_U32_CLASSID]), b1));
+			sprint_tc_classid(rta_getattr_u32(tb[TCA_U32_CLASSID]), b1));
 	} else if (sel && sel->flags&TC_U32_TERMINAL) {
 		fprintf(f, "terminal flowid ??? ");
 	}
 	if (tb[TCA_U32_LINK]) {
 		SPRINT_BUF(b1);
-		fprintf(f, "link %s ", sprint_u32_handle(*(__u32*)RTA_DATA(tb[TCA_U32_LINK]), b1));
+		fprintf(f, "link %s ",
+			sprint_u32_handle(rta_getattr_u32(tb[TCA_U32_LINK]), b1));
 	}
 
 	if (tb[TCA_U32_PCNT]) {
@@ -1018,17 +1244,12 @@ static int u32_print_opt(struct filter_util *qu, FILE *f, struct rtattr *opt, __
 	}
 
 	if (sel) {
-		int i;
-		struct tc_u32_key *key = sel->keys;
 		if (sel->nkeys) {
-			for (i=0; i<sel->nkeys; i++, key++) {
-				fprintf(f, "\n  match %08x/%08x at %s%d",
-					(unsigned int)ntohl(key->val),
-					(unsigned int)ntohl(key->mask),
-					key->offmask ? "nexthdr+" : "",
-					key->off);
+			int i;
+			for (i=0; i<sel->nkeys; i++) {
+				show_keys(f, sel->keys + i);
 				if (show_stats && NULL != pf)
-					fprintf(f, " (success %lld ) ",
+					fprintf(f, " (success %llu ) ",
 						(unsigned long long) pf->kcnts[i]);
 			}
 		}
@@ -1036,7 +1257,9 @@ static int u32_print_opt(struct filter_util *qu, FILE *f, struct rtattr *opt, __
 		if (sel->flags&(TC_U32_VAROFFSET|TC_U32_OFFSET)) {
 			fprintf(f, "\n    offset ");
 			if (sel->flags&TC_U32_VAROFFSET)
-				fprintf(f, "%04x>>%d at %d ", ntohs(sel->offmask), sel->offshift,  sel->offoff);
+				fprintf(f, "%04x>>%d at %d ",
+					ntohs(sel->offmask),
+					sel->offshift,  sel->offoff);
 			if (sel->off)
 				fprintf(f, "plus %d ", sel->off);
 		}
@@ -1055,7 +1278,7 @@ static int u32_print_opt(struct filter_util *qu, FILE *f, struct rtattr *opt, __
 	}
 	if (tb[TCA_U32_INDEV]) {
 		struct rtattr *idev = tb[TCA_U32_INDEV];
-		fprintf(f, "\n  input dev %s\n", (char *) RTA_DATA(idev));
+		fprintf(f, "\n  input dev %s\n", rta_getattr_str(idev));
 	}
 	if (tb[TCA_U32_ACT]) {
 		tc_print_action(f, tb[TCA_U32_ACT]);
